/*
 * Tarea 2
 *	Integrantes:
 * 		- Richard Atuan Toledo
 *		- Felipe Fernández Pino
 */

#include <fcntl.h>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <vector>
#include <sys/wait.h>


typedef struct p {
	short data[160];
} paquete;

using namespace std;

/*
 * Funcion que traspasa los valores de las muestras de audio a vectores en
 * octave.
 */
void traspasarAudios(FILE* sd, char* name_1, const char* name_2, const char* name_3) {
	FILE *file1, *file2, *file3;
	file1 = fopen(name_1, "rb+");
	file2 = fopen(name_2, "rb+");
	file3 = fopen(name_3, "rb+");

	fprintf(sd, "y= [");
	while (!feof(file1)) {
		short data;
		for (int i = 0; i < 160; i++) {
			fread(&data, sizeof(data), 1, file1);
			if (feof(file1)) break;

			fprintf(sd, "%d ", data);
		}
	}
	fprintf(sd, "];\n"); //fflush(sd);

	fprintf(sd, "y2= [");
	while (!feof(file2)) {
		short data;
		for (int i = 0; i < 160; i++) {
			fread(&data, sizeof(data), 1, file2);

			if (feof(file2)) break;

			fprintf(sd, "%d ", data);
		}
	}
	fprintf(sd, "];\n"); //fflush(sd);

	fprintf(sd, "y3= [");
	while (!feof(file3)) {
		short data;
		for (int i = 0; i < 160; i++) {
			fread(&data, sizeof(data), 1, file3);

			if (feof(file3)) break;

			fprintf(sd, "%d ", data);
		}
	}
	fprintf(sd, "];\n"); //fflush(sd);
	fclose(file1);
	fclose(file2);
	fclose(file3);
}

/*
 * Funcion que muestra la grafica de los tres archivos de audio.
 */
void mostrarGrafico(FILE *sd, int contador, float offset) {
	fprintf(sd, "x= [");
	for (int i = 0; i < contador; i++)
		fprintf(sd, "%f ", i*0.000125);
	fprintf(sd, "];\n");

	fprintf(sd, "y2 = y2/max(y);\n");
	fprintf(sd, "y3 = y3/max(y);\n");
	fprintf(sd, "y = y/max(y);\n"); //fflush(sd);
	fprintf(sd, "f2 = figure('Position', [300 100 700 700]);\n");
	fprintf(sd, "set(f2, 'visible', 'on');\n");
	fprintf(sd, "subplot(5,1,1);\n");
	fprintf(sd, "plot(x,y);\n");
	fprintf(sd, "axis([%f %f+0.4 -1 1]);\n", offset, offset);
	fprintf(sd, "xlabel('Tiempo [ms]');\n");
	fprintf(sd, "ylabel('Audio');\n");
	fprintf(sd, "title('Original', 'fontsize', 15);\n");
	fprintf(sd, "subplot(5,1,3);\n");
	fprintf(sd, "plot(x,y2);\n");
	fprintf(sd, "axis([%f %f+0.4 -1 1]);\n", offset, offset);
	fprintf(sd, "xlabel('Tiempo [ms]');\n");
	fprintf(sd, "ylabel('Audio');\n");
	fprintf(sd, "title('Con perdidas', 'fontsize', 15);\n");
	fprintf(sd, "subplot(5,1,5);\n");
	fprintf(sd, "plot(x,y3);\n");
	fprintf(sd, "axis([%f %f+0.4 -1 1]);\n", offset, offset);
	fprintf(sd, "xlabel('Tiempo [ms]');\n");
	fprintf(sd, "ylabel('Audio');\n");
	fprintf(sd, "title('Recuperado', 'fontsize', 15);\n");
	fprintf(sd, "pause;\n");
	fprintf(sd, "\n");
}

int main (int argc, char *argv[]) {
	pid_t pid;
	int pfd[2], pfd2[2];
	int status;
	int seed;
	FILE *original, *conCeros, *recuperado, *sd, *sd2;
	float tasa_de_perdida;
	float offset;
	float e; /* Indice de error */
	bool reproducir = false;
	char c; /* Caracter auxiliar para fscanf */

	if ((argc < 5) || (argc > 6)) {
		printf("Uso:\n");
		printf("\t%s <archivo_de_audio_original>  <tasa_de_pérdida> <offset> <srand> [p] \n", argv[0]);
		exit(1);
	}
	else if (argc == 6) {
		if (!strcmp(argv[5], "p")) {
			reproducir = true;
		}
		else {
			printf("Uso:\n");
			printf("\t%s <archivo_de_audio_original>  <tasa_de_pérdida> <offset> <srand> [p] \n", argv[0]);
			exit(-1);
		}
	}

	/*
	 * Creacion de las dos pipas (envio y recepcion).
	 */
	if ((pipe(pfd) < 0) || (pipe(pfd2) < 0)) {
		perror("pipe");
		exit(1);
	}

	/*
	 * Crear proceso hijo.
	 */
	if ((pid = fork()) < 0) {
		perror("fork");
		exit(1);
	}

	/*
	 * El proceso hijo ejecuta octave.
	 */
	if (pid == 0) {
		int junk;
	
		/*
		 * Enlazamos la entrada estandar con input de pipe1,
		 * y enlazamos la salida estandar con output de pipe2.
		 */
		dup2(pfd[0], 0);
		dup2(pfd2[1], 1);
		close(pfd[1]);  /* cerramos output de pipe1 */
		close(pfd2[0]);  /* cerramos input de pipe2 */
		junk = open("/dev/null", O_WRONLY);
		dup2(junk, 2);  /* Descartamos cualquier mensaje de error de octave */
		execlp("octave", "octave", "-q", (char *) 0);
		perror("exec");
		exit(-1);
	}

	/*
	 * Cerramos el input de pipe1 y cerramos el output de pipe2
	 */
	close(pfd[0]);
	close(pfd2[1]);
	sd = fdopen(pfd[1], "w");
	sd2 = fdopen(pfd2[0], "r");

	/*
	 * Lectura de argumentos
	 */
	original = fopen(argv[1], "rb+");
	tasa_de_perdida = atof(argv[2]);
	offset = atof(argv[3]);
	seed = atoi(argv[4]);

	conCeros = fopen ("conCeros.raw", "wb+");
	recuperado = fopen ("recuperado.raw", "wb+");

	srand(seed);

	paquete prev = {{0}};
	fpos_t pos_i, pos_f, pos_m;

	short interpol_inicial[10], interpol_final[10];
	bool last_was_lost = false;

	/*
	 * Procesamiento de archivos
	 */
	int contador = 0;

	while (!feof(original)) {
		short withZeros, recovered, data, zero = 0;
		int inicio, fin;
		bool loss = static_cast <float> (rand())/RAND_MAX < tasa_de_perdida;

		if (loss) {
			/*
			 * Almacenamiento de datos para interpolacion inicial y final.
			 */
			interpol_inicial[0] = prev.data[155];
			interpol_inicial[1] = prev.data[156];
			interpol_inicial[2] = prev.data[157];
			interpol_inicial[3] = prev.data[158];
			interpol_inicial[4] = prev.data[159];
			interpol_final[0] = prev.data[145];
			interpol_final[1] = prev.data[146];
			interpol_final[2] = prev.data[147];
			interpol_final[3] = prev.data[148];
			interpol_final[4] = prev.data[149];
			fgetpos(recuperado, &pos_i);
		}
		inicio = contador;

		for (int i = 0; i < 160; i++) {
			/*
			 * Escritura de datos en los archivos.
			 */
			fread((char *) &data, sizeof(data), 1, original);

			if (feof(original)) break;

			if (loss)
				withZeros = zero;
			else
				withZeros = data;

			fwrite((char*) &withZeros, sizeof(withZeros), 1, conCeros);

			if (loss)
				recovered = prev.data[i];
			else
				recovered = data;

			if ((i >= 9) && (i<= 13) && loss) {
				interpol_inicial[i-4] = recovered;
			}

			if ((i >= 0) && (i <= 4) && last_was_lost) {
				interpol_final[i+5] = recovered;
			}

			if ((loss) && (i == 150))
				fgetpos(recuperado, &pos_m);

			fwrite((char*) &recovered, sizeof(recovered), 1, recuperado);

			prev.data[i] = withZeros;
			contador++;
		}
		fin = contador;

		/*
		 * Interpolacion final.
		 */
		if (last_was_lost && !loss) {
			fgetpos(recuperado, &pos_f);
			fprintf(sd, "yprev2= [");
			for (int i = 0; i < 10; i++) {
				fprintf(sd, "%d ", interpol_final[i]);
			}
			fprintf(sd, "];\n");
			fprintf(sd, "xprev2= [");
			for (int i = inicio-15; i < inicio-10; i++) {
				fprintf(sd, "%f ", i*0.000125);
			}
			for (int i = inicio; i < inicio+5; i++) {
				fprintf(sd, "%f ", i*0.000125);
			}
			fprintf(sd, "];\n");
			fprintf(sd, "p = polyfit(xprev2, yprev2, 10);\n");

			fsetpos(recuperado, &pos_m);	
			short interpol;
			float interpol_temp;
			for (int j = 0; j < 10; j++) {
				fprintf(sd, "polyval(p,%f)\n", (inicio+j-10)*0.000125);
				fflush(sd);
				fscanf(sd2, "ans = %f", &interpol_temp);
				fscanf(sd2, "%c", &c);
				interpol = (short) interpol_temp;
				fwrite((char *) &interpol, sizeof(interpol), 1, recuperado);
			}
			fsetpos(recuperado, &pos_f);
		}

		/*
		 * Interpolacion inicial.
		 */
		if (loss) {
			fgetpos(recuperado, &pos_f);
			fprintf(sd, "yprev= [");
			for (int i = 0; i < 10; i++) {
				fprintf(sd, "%d ", interpol_inicial[i]);
			}
			fprintf(sd, "];\n");
			fprintf(sd, "xprev= [");
			for (int i = inicio-5; i < inicio; i++) {
				fprintf(sd, "%f ", i*0.000125);
			}
			for (int i = inicio+9; i < inicio+14; i++) {
				fprintf(sd, "%f ", i*0.000125);
			}
			fprintf(sd, "];\n");
			fprintf(sd, "p = polyfit(xprev, yprev, 10);\n");

			fsetpos(recuperado, &pos_i);	
			short interpol;
			float interpol_temp;
		
			for (int j = 0; j < 10; j++) {
				fprintf(sd, "polyval(p,%f)\n", (inicio+j)*0.000125);
				fflush(sd);
				fscanf(sd2, "ans = %f", &interpol_temp);
				fscanf(sd2, "%c", &c);
				interpol = (short) interpol_temp;
				fwrite((char *) &interpol, sizeof(interpol), 1, recuperado);
			}
			fsetpos(recuperado, &pos_f);
			last_was_lost = true;
		}
		else 
			last_was_lost = false;
	}

	fclose(original);
	fclose(conCeros);
	fclose(recuperado);

    system("clear");
    

	if (reproducir) {
		char nombre[100];
		sprintf(nombre, "aplay --format=S16_LE -t raw %s", argv[1]);
		system(nombre);
		system("aplay --format=S16_LE -t raw conCeros.raw");
		system("aplay --format=S16_LE -t raw recuperado.raw");
		
	}
	printf("Cargando Datos...\n");
	
	traspasarAudios(sd, argv[1], "conCeros.raw", "recuperado.raw");
    /*
	 * Calculo del indice de error.
	 */
	fprintf(sd, "e = sum((y2-y).^2)\n");
	fflush(sd);
	fscanf(sd2, "e = %f", &e);
	fscanf(sd2, "%c", &c);
	printf("Indice de error = %f\n", e);
	
	printf("Cargando Graficos...\n");
	mostrarGrafico(sd, contador, offset);

	printf("Presione cualquier tecla para finalizar.\n");
    	
	fprintf(sd, "pause();\n");
	fflush(sd);
	system ("/bin/stty raw");
	char u;
	if ((u = getchar())) {
		fprintf(sd, "\n");
	}
	system ("/bin/stty cooked");
        fprintf(sd, "\n exit\n"); fflush(sd);
	/*
	 * Wait for the child to exit and
	 * close de pipe.
	 */
	waitpid(pid, &status, 0);
	fclose(sd);
}
